#ifndef MODEL_H
#define MODEL_H

#include <vector>
#include <string>

class Mesh;
class Image;
class Texture;

class Model
{
protected:
    struct TextureImg
    {
        std::string name;
        Texture* tex;
        Image* img;
        
        TextureImg(const std::string n, Image* i) :
            name(n), tex(NULL), img(i) {}
    };
    
    std::vector<Mesh*> m_meshes;
    std::vector<TextureImg> m_textures;

public:
    ~Model();
    
    /**
     * @brief addMesh adds a mesh to the mesh list
     */
    void addMesh(Mesh *mesh) { m_meshes.push_back(mesh); }
    
    /**
     * @brief addImage adds an image to the texture pack
     */
    void addImage(const std::string &name, Image* img)
    {
        m_textures.push_back(TextureImg(name, img));
    }
    
    /**
     * @brief getMeshes returns the mesh array
     */
    const std::vector<Mesh*>& getMeshes() { return m_meshes; }
    
    /**
     * @brief getImage returns the image which has the specified name, or NULL if there are no images of this name.
     */
    Image* getImage(const std::string &name);
    
    /**
     * @brief getTexture returns the texture which has the specified name, or NULL if there are no images of this name.
     */
    Texture* getTexture(const std::string &name);
    
    /**
     * @brief destroy frees Mesh and Image memory in the RAM, but it stays allocated in GPU memory.
     */
    void destroy();
    
    /**
     * @brief initGL uploads image and mesh data to GPU memory
     */
    void initGL();
    
    /**
     * @brief destroyGL frees the meshes and the textures from GPU memory
     */
    void destroyGL();
    
    /**
     * @brief save saves the model on a file
     * @param filename is the name of the resulting file
     * @param texturesPath is the existing folder where the textures will be saved
     * @return true if the save succeeded
     */
    bool save(const std::string &filename, const std::string &texturesPath);
    
    /**
     * @brief load loads the model from a file
     * @param filename is the name of the target file
     * @param texturesPath is the existing folder where the textures will be gathered
     * @return the model if the loading succeeded, or NULL if it failed
     */
    static Model* load(const std::string &filename, const std::string &texturesPath);
};

#endif // MODEL_H
